from __future__ import annotations

import json
import time
from base64 import b64encode
from datetime import timedelta
from io import StringIO
from ipaddress import IPv4Address, IPv6Address, ip_network
from typing import List

import click
from nacl.exceptions import BadSignatureError
from nacl.signing import SigningKey, VerifyKey
from schema import And, Optional, Regex, Schema, Use

from .common import (
    Bug,
    ConsistencyError,
    Flexibool,
    Length,
    attrdict,
    b64_bytes,
    comma_separated_IPs,
    comma_separated_Nets,
    format_byte_stats,
    int_range,
    organize_dbus_if_active,
    queryable,
    schemadict,
    schemattrdict,
    serializable,
    yamlrepr,
    yamlrepr_hl,
)
from .engine import Result
from .notclick import (
    DualUse,
    blue,
    bold,
    echo_maybepager,
    green,
    red,
    shell_complete_helper,
    yellow,
)

_qrcode = None


@DualUse.object()
class Descriptor(schemattrdict, serializable):

    """
    Descriptors are the objects which are communicated between between publish
    and discover (via mDNS) and between discover and organize (via strings).

    This ridiculous test uses the fact that this is commandline-accessible:

    >>> x = Descriptor.cli.main(standalone_mode=False, args=tuple("--addrs 10.168.128.160 --c "
    ... "vdDpRSGtsqvui8dox0iBq0SSp/zXSEU2dx5s5x+qcquSp0oIWgDuqJw50e9wrIuGub+SXzU0s5EIRH49QmNYDw=="
    ... " --dt 86400 --e false --hostname wg-mdns-test3.local.  --pk EqcQ5gYxzGtzg7B4xi83kLyfuSMp8K"
    ... "v3cmAJMs12nDM= --port 5354 --s T6htsKgwCp5MAXjPiWxtVkccg+K2CePsVa7uyUgxE2ouYKXg2qNL+0ut3s"
    ... "SbVTYjzFGZSCO6n80SRaR+BIeOCg== --vf 1606276812 --vk 90Y5JGEjoklSDw51ffoHYXhWs49TTnCQ/D5yB"
    ... "buf3Zg= valid".split()))
    True

    For testing purposes, one could run the same command on the commandline like so:
    vula -d peer.Descriptor --addrs 10.168.128.160 --c
    vdDpRSGtsqvui8dox0iBq0SSp/zXSEU2dx5s5x+qcquSp0oIWgDuqJw50e9wrIuGub+SXzU0s5EIRH49QmNYDw== --dt
    86400 --e false --hostname wg-mdns-test3.local.  --pk
    EqcQ5gYxzGtzg7B4xi83kLyfuSMp8Kv3cmAJMs12nDM= --port 5354 --r '' --s
    T6htsKgwCp5MAXjPiWxtVkccg+K2CePsVa7uyUgxE2ouYKXg2qNL+0ut3sSbVTYjzFGZSCO6n80SRaR+BIeOCg== --vf
    1606276812 --vk 90Y5JGEjoklSDw51ffoHYXhWs49TTnCQ/D5yBbuf3Zg= valid

    ... IPv6
    Unfortunately, this test can't be adapted to IPv6 as it's not possible to get a valid signature.

    For testing purposes, one could run the same command on the commandline like so:
    vula -d peer.Descriptor --addrs fe80::377b:d17:9b74:1b91 --c
    vdDpRSGtsqvui8dox0iBq0SSp/zXSEU2dx5s5x+qcquSp0oIWgDuqJw50e9wrIuGub+SXzU0s5EIRH49QmNYDw== --dt
    86400 --e false --hostname wg-mdns-test3.local.  --pk
    EqcQ5gYxzGtzg7B4xi83kLyfuSMp8Kv3cmAJMs12nDM= --port 5354 --r '' --s
    T6htsKgwCp5MAXjPiWxtVkccg+K2CePsVa7uyUgxE2ouYKXg2qNL+0ut3sSbVTYjzFGZSCO6n80SRaR+BIeOCg== --vf
    1606276812 --vk 90Y5JGEjoklSDw51ffoHYXhWs49TTnCQ/D5yBbuf3Zg= valid

    ... but this is of course not the correct way to use this object.
    """

    schema = Schema(
        {
            'addrs': Use(
                comma_separated_IPs
            ),  # This is a list of endpoints by preference and is also used for AllowedIPs;
            # fixme in the future
            'pk': b64_bytes.with_len(32),
            'c': b64_bytes.with_len(64),
            'hostname': And(
                str,
                # fixme: this should also ensure labels don't
                #  start or end with a dash, as per RFC1123
                Regex('^[a-zA-Z0-9.-]+$'),
                lambda name: 0 < len(name) < 256,
            ),
            'port': int_range(1, 65535),
            'vk': b64_bytes.with_len(32),
            'dt': Use(int),
            'vf': Use(int),
            'r': Use(
                comma_separated_Nets
            ),  # AllowedIPs may be set from this in the future; fixme
            'e': Flexibool,
            Optional('s'): b64_bytes.with_len(64),
        }
    )

    default = dict(r="")

    @classmethod
    def parse(cls, desc: str) -> Descriptor:
        """
        Parse the *descriptor* string line into a dictionary-like object. Carefully.

        This relies on the schema to coerce values into the right types.
        >>> d = Descriptor.parse("addrs=192.168.2.106;"
        ...            "pk=/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg=;"
        ...            "c=SbS/v8aTX2Ti3E1jp4T7d8lHoW4v5iRcBcdQ6tSv5bzCuYg9h56A4YFjkKv9aFA+"
        ...            "A7u0yrqYhTdpuAWuBgL2BQ==;"
        ...            "hostname=bubu.local;"
        ...            "port=5354;"
        ...            "vk=/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg=;"
        ...            "dt=86400;"
        ...            "vf=1636045801;"
        ...            "e=false")
        >>> d
        {'r': <comma_separated_Nets('')>, 'addrs': <comma_separated_IPs('192.168.2.106')>, 'pk': \
<b64:/O6SlK...(32)>, 'c': <b64:SbS/v8...(64)>, 'hostname': 'bubu.local', 'port': 5354, \
'vk': <b64:/O6SlK...(32)>, 'dt': 86400, 'vf': 1636045801, 'e': 0}
        >>> str(d.addrs)
        '192.168.2.106'
        >>> str(d.pk)
        '/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg='
        >>> str(d.c)
        'SbS/v8aTX2Ti3E1jp4T7d8lHoW4v5iRcBcdQ6tSv5bzCuYg9h56A4YFjkKv9aFA+A7u0yrqYhTdpuAWuBgL2BQ=='
        >>> d.hostname
        'bubu.local'
        >>> d.port
        5354
        >>> str(d.vk)
        '/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg='
        >>> d.dt
        86400
        >>> d.vf
        1636045801
        >>> d.e
        0

        This relies on the schema to coerce values into the right types.
        >>> d = Descriptor.parse("addrs=fe80::377b:d17:9b74:1b91;"
        ...            "pk=/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg=;"
        ...            "c=SbS/v8aTX2Ti3E1jp4T7d8lHoW4v5iRcBcdQ6tSv5bzCuYg9h56A4YFjkKv9aFA+"
        ...            "A7u0yrqYhTdpuAWuBgL2BQ==;"
        ...            "hostname=bubu.local;"
        ...            "port=5354;"
        ...            "vk=/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg=;"
        ...            "dt=86400;"
        ...            "vf=1636045801;"
        ...            "e=false")
        >>> d
        {'r': <comma_separated_Nets('')>, 'addrs': <comma_separated_IPs('fe80::377b:d17:9b74:1b91')>, 'pk': \
<b64:/O6SlK...(32)>, 'c': <b64:SbS/v8...(64)>, 'hostname': 'bubu.local', 'port': 5354, \
'vk': <b64:/O6SlK...(32)>, 'dt': 86400, 'vf': 1636045801, 'e': 0}
        >>> str(d.addrs)
        'fe80::377b:d17:9b74:1b91'
        >>> str(d.pk)
        '/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg='
        >>> str(d.c)
        'SbS/v8aTX2Ti3E1jp4T7d8lHoW4v5iRcBcdQ6tSv5bzCuYg9h56A4YFjkKv9aFA+A7u0yrqYhTdpuAWuBgL2BQ=='
        >>> d.hostname
        'bubu.local'
        >>> d.port
        5354
        >>> str(d.vk)
        '/O6SlKeiGWHyK0VpA1V5emqseJqWdDs/J8Mu6SGEQHg='
        >>> d.dt
        86400
        >>> d.vf
        1636045801
        >>> d.e
        0
        """
        try:
            split_desc: List = desc.split(";")
            dir_desc: dict = dict(
                val.split("=", maxsplit=1)
                for val in [kv.strip() for kv in split_desc]
                if len(val) > 1
            )
            descriptor: Descriptor = cls(dir_desc)
        except ValueError:
            raise
            # log.info("Unable to parse descriptor: %s (%r)", error, descriptor)
            return None
        return descriptor

    def _build_sig_buf(self: Descriptor) -> bytes:
        return " ".join(
            "%s=%s;" % (k, v) for k, v in sorted(self.items()) if k != 's'
        ).encode()

    def __str__(self):
        """
        Return the number or IP as a string

        >>> ip = comma_separated_IPs('192.168.13.37')
        >>> ip.__str__()
        '192.168.13.37'

        >>> ip = comma_separated_IPs('fe80::377b:d17:9b74:1b91')
        >>> ip.__str__()
        'fe80::377b:d17:9b74:1b91'
        """
        return " ".join("%s=%s;" % kv for kv in sorted(self.items()))

    @property
    def id(self):
        """
        This returns the peer ID (aka the verify key, base64-encoded)
        >>> desc_s = (
        ... "addrs=192.168.6.9;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;"
        ... "dt=86400;e=0;hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=5354;"
        ... "vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.id
        'Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M='

        >>> desc_s = (
        ... "addrs=fe80::377b:d17:9b74:1b91;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;"
        ... "dt=86400;e=0;hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;"
        ... "port=5354;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.id
        'Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M='

        """
        return str(self.vk)

    @property
    def ephemeral(self):
        """
        >>> desc_s = (
        ... "addrs=192.168.6.9;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;"
        ... "dt=86400;e=0;hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;"
        ... "port=5354;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.ephemeral
        0

        >>> desc_s = (
        ... "addrs=fe80::377b:d17:9b74:1b91;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;"
        ... "dt=86400;e=0;hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;"
        ... "port=5354;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.ephemeral
        0
        """
        return self.e

    @property
    def valid(self):
        # XXX: placeholder until this is merged with other class with signature validation etc
        # ^^^
        return self.verify_signature()

    @property
    def IPv4addrs(self):
        """
        >>> desc_s = (
        ... "addrs=2001:0db8:85a3:0000:0000:8a2e:0370:7334;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;"
        ... "dt=86400;e=0;hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;"
        ... "port=5354;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.IPv4addrs
        []
        >>> desc_s = (
        ... "addrs=192.168.6.9;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;dt=86400;e=0;"
        ... "hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=5354;vf=1601388653;"
        ... "vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.IPv4addrs
        [IPv4Address('192.168.6.9')]
        """
        return [ip for ip in self.addrs if isinstance(ip, IPv4Address)]

    @property
    def IPv6addrs(self):
        """
        >>> desc_s = (
        ... "addrs=2001:0db8:85a3:0000:0000:8a2e:0370:7334;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;"
        ... "dt=86400;e=0;hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=5354;"
        ... "vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.IPv6addrs
        [IPv6Address('2001:db8:85a3::8a2e:370:7334')]
        >>> desc_s = (
        ... "addrs=192.168.6.9;"
        ... "c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;"
        ... "dt=86400;e=0;hostname=alice.local;pk=QkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;"
        ... "port=5354;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> d = Descriptor.parse(desc_s)
        >>> d.IPv6addrs
        []
        """
        return [ip for ip in self.addrs if isinstance(ip, IPv6Address)]

    def sign(self, seed) -> Descriptor:
        signing_key: SigningKey = SigningKey(seed=seed)
        buf_to_sign: bytes = self._build_sig_buf()
        sig: bytes = signing_key.sign(buf_to_sign).signature
        descriptor = Descriptor(self, s=b64encode(sig).decode())
        return descriptor

    def verify_signature(self) -> bool:
        """
        Verifies the signature.
        Returns true if valid, false if invalid.

        Valid Signature:
        >>> desc_str = (
        ... "addrs=10.215.167.50; c=KNDxDMgmkH8Poa7TJBlIZrvTnQBN5w10gYlyY5"
        ... "MfvkA7Eu12IhpheCdJzWIwap4PE5Ryv3PzvU4ikrEY6oXJNw==; dt=86400; e=0; "
        ... "hostname=wg-mdns-test1.local.; pk=y9bQa4DAj4NT5lh8PffyAbXNbYCkxczMKLk/r"
        ... "tP4CVY=; port=5354; r=; s=YJqLUPrI8G/IfA1wIbW2z5p0EtYcDFh4gxCjP5czMK2wi"
        ... "GRgZdeBibs6shDoRusfHtSy+4m/Z9Jfhul+amQYAQ==; vf=1605737957; vk=XGQErb1N"
        ... "Jmg4dMLZK7hXfhRahgZ6ix/oP3+BTq2+Dy8=;")
        >>> desc = Descriptor.parse(desc_str)
        >>> assert desc.verify_signature() is True

        Invalid Signature:
        >>> desc = Descriptor(desc, port=desc['port'] + 1)
        >>> assert desc.verify_signature() is False
        """
        verify_public_key = VerifyKey(self.vk)
        sig = self.s
        buf_to_verify: bytes = self._build_sig_buf()
        try:
            verify_public_key.verify(buf_to_verify, sig)
            return True
        except BadSignatureError:
            return False

    def make_peer(self: Descriptor, **kwargs) -> Peer:
        peer = dict(
            descriptor=self,
            petname="",
            pinned=False,
            enabled=True,
            verified=False,
            nicknames={self.hostname: True},
            IPv4addrs={ip: True for ip in self.IPv4addrs},
            IPv6addrs={ip: True for ip in self.IPv6addrs},
            use_as_gateway=False,
        )
        peer.update(**kwargs)
        return Peer(peer)

    @property
    def qr_code(self):
        """
        This generates a QR-Code and prints it.
        The data that is encoded within the QR-Code is
        given by this functions parameter.

        It returns a String.
        """
        global _qrcode
        if _qrcode is None:
            import qrcode as _qrcode
        sio = StringIO()
        qr = _qrcode.QRCode()
        qr.add_data(data="local.vula:desc:" + str(self))
        qr.print_ascii(out=sio)
        sio.seek(0)
        return sio.read()


class Peer(schemattrdict):
    schema = Schema(
        And(
            {
                'descriptor': Use(Descriptor),
                'petname': str,
                'nicknames': {Optional(str): Flexibool},
                'IPv4addrs': {Optional(Use(IPv4Address)): Flexibool},
                'IPv6addrs': {Optional(Use(IPv6Address)): Flexibool},
                'enabled': Flexibool,
                'verified': Flexibool,
                'pinned': Flexibool,
                Optional('use_as_gateway'): Flexibool,
                Optional('_allow_unsigned_descriptor'): Flexibool,
            },
            # lambda peer:
            # peer['descriptor'].verify_signature() or peer.get('_allow_unsigned_descriptor')
        )
    )

    # Peers are currently only instantiated by Descriptor.make_peer, default is
    # there.
    default = None

    @property
    def id(self):
        "This returns the peer ID (aka the verify key, as a base64 string)"
        return self.descriptor.id

    @property
    def wg_pk(self):
        "WireGuard public key"
        return self.descriptor.pk

    @property
    def name(self):
        """
        This returns the best name, for display purposes.

        It is either the petname, or the last name announced, or, if the last
        name announced is disabled, another enabled name.
        >>> desc_string = ("addrs=192.168.6.9;c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUF"
        ... f"BQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;dt=86400;e=0;hostname=george.local;pk=QkJCQkJCQ"
        ... f"kJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=123;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0N"
        ... f"DQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> descriptor = Descriptor.parse(desc_string)
        >>> Peer(dict(descriptor=descriptor,petname="george",pinned=False,enabled=True,
        ... verified=False,nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip
        ...  in descriptor.IPv4addrs},IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).name
        'george'
        >>> Peer(dict(descriptor=descriptor,petname="",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip in descriptor.IPv4addrs
        ... },IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).name
        'george.local'
        >>> Peer(dict(descriptor=descriptor,petname="",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: False, "schnubbi": True},IPv4addrs={ip: True for ip
        ...  in descriptor.IPv4addrs},IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).name
        'schnubbi'

        >>> desc_string = ("addrs=fe80::377b:d17:9b74:1b91;c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUF"
        ... f"BQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;dt=86400;e=0;hostname=george.local;pk=QkJCQkJCQ"
        ... f"kJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=123;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0N"
        ... f"DQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> descriptor = Descriptor.parse(desc_string)
        >>> Peer(dict(descriptor=descriptor,petname="george",pinned=False,enabled=True,
        ... verified=False,nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip
        ...  in descriptor.IPv4addrs},IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).name
        'george'
        """
        latest = self.descriptor.hostname
        return self.get('petname') or (
            latest
            if self.nicknames[self.descriptor.hostname]
            else self.enabled_names[0]
            if self.enabled_names
            else "<unnamed>"
        )

    @property
    def other_names(self):
        """
        Returns a sorted list of all names other than self.name

        >>> desc_string = ("addrs=192.168.6.9;c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUF"
        ... f"BQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;dt=86400;e=0;hostname=george.local;pk=QkJCQkJCQ"
        ... f"kJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=123;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0N"
        ... f"DQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> descriptor = Descriptor.parse(desc_string)
        >>> Peer(dict(descriptor=descriptor,petname="",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip in descriptor.IPv4addrs
        ... },IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).other_names
        []
        >>> Peer(dict(descriptor=descriptor,petname="george",pinned=False,enabled=True,
        ... verified=False,nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip
        ...  in descriptor.IPv4addrs},IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).other_names
        ['george.local']
        >>> Peer(dict(descriptor=descriptor,petname="george",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: False, "schnubbi": True},IPv4addrs={ip: True for ip
        ...  in descriptor.IPv4addrs},IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).other_names
        ['george.local', 'schnubbi']

        >>> desc_string = ("addrs=fe80::377b:d17:9b74:1b91;c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUF"
        ... f"BQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;dt=86400;e=0;hostname=george.local;pk=QkJCQkJCQ"
        ... f"kJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=123;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0N"
        ... f"DQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> descriptor = Descriptor.parse(desc_string)
        >>> Peer(dict(descriptor=descriptor,petname="",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip in descriptor.IPv4addrs
        ... },IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).other_names
        []
        >>> Peer(dict(descriptor=descriptor,petname="george",pinned=False,enabled=True,
        ... verified=False,nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip
        ...  in descriptor.IPv4addrs},IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).other_names
        ['george.local']
        >>> Peer(dict(descriptor=descriptor,petname="george",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: False, "schnubbi": True},IPv4addrs={ip: True for ip
        ...  in descriptor.IPv4addrs},IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).other_names
        ['george.local', 'schnubbi']
        """
        return list(sorted(set(self.nicknames) - set([self.name])))

    @property
    def name_and_id(self):
        """
        Returns the name and id of the peer.

        >>> desc_string = ("addrs=192.168.6.9;c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUF"
        ... f"BQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;dt=86400;e=0;hostname=george.local;pk=QkJCQkJCQ"
        ... f"kJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=123;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0N"
        ... f"DQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> descriptor = Descriptor.parse(desc_string)
        >>> Peer(dict(descriptor=descriptor,petname="",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip in descriptor.IPv4addrs
        ... },IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).name_and_id
        'george.local (Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=)'

        >>> desc_string = ("addrs=fe80::377b:d17:9b74:1b91;c=QUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUF"
        ... f"BQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQQ==;dt=86400;e=0;hostname=george.local;pk=QkJCQkJCQ"
        ... f"kJCQkJCQkJCQkJCQkJCQkJCQkJCQkJCQkI=;port=123;vf=1601388653;vk=Q0NDQ0NDQ0NDQ0NDQ0NDQ0N"
        ... f"DQ0NDQ0NDQ0NDQ0NDQ0M=")
        >>> descriptor = Descriptor.parse(desc_string)
        >>> Peer(dict(descriptor=descriptor,petname="",pinned=False,enabled=True,verified=False,
        ... nicknames={descriptor.hostname: True},IPv4addrs={ip: True for ip in descriptor.IPv4addrs
        ... },IPv6addrs={ip: True for ip in descriptor.IPv6addrs})).name_and_id
        'george.local (Q0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0NDQ0M=)'
        """
        return "%s (%s)" % (self.name, self.id)

    @property
    def enabled_names(self):
        return sorted(
            set(
                ([self.petname] if self.petname else [])
                + [n for n in self.nicknames if self.nicknames[n]]
            )
        )

    @property
    def enabled_ips(self):
        return [ip for ip, on in self.IPv4addrs.items() if on] + [
            ip for ip, on in self.IPv6addrs.items() if on
        ]

    @property
    def disabled_ips(self):
        return [ip for ip, on in self.IPv4addrs.items() if not on] + [
            ip for ip, on in self.IPv6addrs.items() if not on
        ]

    @property
    def enabled_ips_str(self):
        return [str(ip) for ip in self.enabled_ips]

    @property
    def endpoint_addr(self):
        return self.descriptor.addrs[0]  # XXX use better "best ip" logic,
        # allowing for disabled IPs, etc

    @property
    def routes(self):
        res = [
            ip_network("%s/%s" % (ip, ip.max_prefixlen))
            for ip in self.enabled_ips
        ]
        return res

    @property
    def allowed_ips(self):
        res = [
            ip_network("%s/%s" % (ip, ip.max_prefixlen))
            for ip in self.enabled_ips
        ]
        if self.use_as_gateway:
            res.append(ip_network("0.0.0.0/0"))  # FIXME: ipv6
        return res

    @property
    def endpoint(self):
        return "%s:%s" % (self.endpoint_addr, self.descriptor.port)

    def wg_config(self: Peer, ctidh_psk):
        return attrdict(
            public_key=str(self.descriptor.pk),
            endpoint_addr=str(self.endpoint_addr),
            endpoint_port=self.descriptor.port,
            allowed_ips=list(map(str, self.allowed_ips)),
            preshared_key=ctidh_psk,
        )

    def show(self, stats=None):
        green_or_yellow = (
            green if self.pinned and self.verified and self.enabled else yellow
        )
        return "\n  ".join(
            bold(label) + ': ' + str(value) if label else str(value)
            for label, value in {
                green_or_yellow('peer'): green_or_yellow(self.name),
                'id': self.id,
                red('warning'): (
                    None if stats else red('wireguard peer is not configured')
                ),
                'other names': ", ".join(self.other_names),
                'status': " ".join(
                    filter(
                        None,
                        [
                            (red("disabled"), green("enabled"))[self.enabled],
                            (yellow("unpinned"), green("pinned"))[self.pinned],
                            (
                                (red if self.pinned else yellow)("unverified"),
                                green("verified"),
                            )[self.verified],
                            ('', bold(blue("gateway")))[self.use_as_gateway],
                        ],
                    )
                ),
                'endpoint': self.endpoint,
                'allowed ips': ", ".join(map(str, self.allowed_ips)),
                'disabled ips': ", ".join(map(str, self.disabled_ips)),
                'latest signature': (
                    str(
                        timedelta(
                            seconds=int(time.time() - self.descriptor.vf)
                        ),
                    )
                    + ' ago'
                ),
                'latest handshake': (
                    str(
                        timedelta(
                            seconds=int(
                                time.time() - stats['latest_handshake']
                            )
                        ),
                    )
                    + ' ago'
                    if stats and stats.get('latest_handshake')
                    else yellow('none')
                ),
                'transfer': (
                    stats
                    and sum(stats.values())
                    and "{rx_bytes} received, {tx_bytes} sent".format(
                        **format_byte_stats(stats)
                    )
                ),
                'wg pubkey': self.descriptor.pk,
            }.items()
            if value not in (None, False, '')
        )


class Peers(yamlrepr, queryable, schemadict):

    """
    A dictionary of peers. Note that, despite being the home of the conflict
    detection code, a Peers object can be valid (from a schema standpoint) even
    while containing conflicts. The schema of a SystemState object uses this
    conflict detection code to ensure that the peers object within a
    SystemState does not contain conflicts.
    """

    schema = Schema(
        {
            Optional(And(str, Length(44))): Use(Peer),
        },
    )

    def with_hostname(self: Peers, name: str):
        res = self.limit(enabled=True).by('nicknames').get(name, [])
        if len(res) > 1:
            raise Bug(
                # this should not be possible, as both the state logic and
                # schema prevent it
                "BUG: multiple hosts with the same name: %r"
                % (res,)
            )
        if len(res) < 1:
            raise KeyError("Hostname %r not found." % (name,))
        return res[0]

    def with_ip(self, ip):
        # IPv6 analysis: not ipv6 ready
        # Please enhance this function to support ipv6
        ip = IPv4Address(ip)
        res = self.limit(enabled=True).by('IPv4addrs').get(ip, [])
        if len(res) > 1:
            raise ConsistencyError(
                # this should also not be possible, because the state logic
                # prevents it (though the schema doesn't yet)
                "BUG: multiple hosts with the same IP: %r"
                % (res,)
            )
        if len(res) < 1:
            raise ValueError("No peer with IP %r" % (ip,))
        return res[0]

    @property
    def conflicts(self):
        "returns comma-separated list of colliding peer ids"
        enabled_gws = list(
            self.limit(use_as_gateway=True, enabled=True).values()
        )
        res = ",".join(
            [
                peer.id
                for peer in self.limit(enabled=True).values()
                if self.conflicts_for_descriptor(peer.descriptor)
            ]
            + [peer.id for peer in enabled_gws if len(enabled_gws) > 1]
        )
        return res

    def conflicts_for_descriptor(self, desc):
        """
        Returns list of enabled peers a descriptor has a conflicting wg_pk,
        hostname, or IP address with (ignoring itself).
        """
        return list(
            {
                conflict.id: conflict
                for conflict in self.limit(enabled=True)
                .by('enabled_names')
                .get(desc.hostname, [])
                + self.limit(enabled=True).by('wg_pk').get(desc.pk, [])
                + sum(
                    (
                        self.limit(enabled=True).by('IPv4addrs').get(ip, [])
                        for ip in desc.IPv4addrs
                    ),
                    [],
                )
                + sum(
                    (
                        self.limit(enabled=True).by('IPv6addrs').get(ip, [])
                        for ip in desc.IPv6addrs
                    ),
                    [],
                )
                if conflict.id != desc.id
            }.values()
        )

    def query(self, query):
        """
        Returns peer by vk, hostname, or IP. None if no match.
        """
        peer = (
            self.by('id').get(query)
            or self.limit(enabled=True).by('enabled_names').get(query)
            or self.limit(enabled=True).by('enabled_ips_str').get(query)
        )
        if peer:
            assert len(peer) == 1, ("this should not be possible:", peer)
            return peer[0]
        else:
            return None


def _ac_get_peer_ids(ctx, args, incomplete):
    organize = (
        ctx.meta.get('Organize', {}).get('magic_instance')
        or organize_dbus_if_active()
    )
    return organize.peer_ids('all')


@DualUse.object(
    short_help="View and modify peer information",
    invoke_without_command=True,
)
@click.pass_context
class PeerCommands(object):
    """
    Commands to view and modify peer information

    When "peer" is the top-level command, its subcommands communicate with
    organize via dbus. This is the normal way to use these commands (eg, you
    can run "vula peer show").

    At this point in development, "peer" may also be run as a subcommand of the
    "organize" command, in which case it will instantiate and operate on the
    organize object directly. In the case of the "set" and "remove" peer
    subcommands this only makes sense when there isn't an organize daemon
    process running (as the daemon process will overwrite the state file
    written by the "organize peer ..." command process without ever reading its
    contents). Note that we do not currently check to ensure that there is not
    an organize daemon running when "organize peer ..." commands are run.
    """

    def __init__(self, ctx):
        self.organize = (
            ctx.meta.get('Organize', {}).get('magic_instance')
            or organize_dbus_if_active()
        )

        if ctx.invoked_subcommand is None:
            self.show()

    @DualUse.method(short_help="Show peer information")
    @click.argument('peers', type=str, nargs=-1)
    @click.option(
        '-D',
        '--descriptor',
        is_flag=True,
        help="Print peer descriptor(s) instead of status",
    )
    @click.option('-q', '--qrcode', is_flag=True, help="Print ANSI qrcode(s)")
    @click.option(
        '-a',
        '--all',
        'which',
        flag_value='all',
        help="Show both enabled and disabled peers",
    )
    @click.option(
        '-d',
        '--disabled',
        'which',
        flag_value='disabled',
        help="Show only disabled peers",
    )
    @click.option(
        '-e',
        '--enabled',
        'which',
        flag_value='enabled',
        help="Show only enabled peers",
    )
    def show(self, peers=(), which=None, descriptor=False, qrcode=False):
        """
        Show peer information

        With no arguments, all enabled peers are shown.

        Peer arguments can be specified as ID, name, or IP, unless options are also
        specified, in which case arguments must (currently) be IDs.
        """

        available = self.organize.peer_ids(which if which else 'enabled')

        queries = (
            available
            if peers == ()
            else [peer for peer in peers if peer in available]
            if which
            else peers
        )

        res = []
        for query in queries:
            if qrcode:
                d = self.organize.peer_descriptor(query)
                desc = Descriptor.parse(d)
                if descriptor:
                    res.append(d)
                else:
                    res.append("{vk} {hostname} {addrs}".format(**desc))
                res.append(desc.qr_code)
            elif descriptor:
                res.append(self.organize.peer_descriptor(query))
            else:
                res.append(self.organize.show_peer(query))

        echo_maybepager(("\n" if descriptor else "\n\n").join(res))

    @DualUse.method(short_help="Import peer descriptors", name='import')
    @click.argument('file', type=click.File(), default='-')
    def import_(self, file):
        """
        Import peer descriptors

        Reads from standard input if a file is not specified.

        Prints the result of processing each descriptor.

        Can consume the output of "vula peer show --descriptor" from another
        system, or the output of "vula discover --no-dbus --interface eth0".
        """
        while True:
            line = file.readline()
            line = line.split(':')[
                -1
            ].strip()  # strip timestamp from discover logs, if present
            if not line:
                break
            line = line.strip()
            res = self.organize.process_descriptor_string(line)
            if res:
                click.echo(yamlrepr_hl.from_yaml(res))
            else:
                click.echo("process_descriptor_string returned None 🤔")

    @DualUse.object(short_help="Add and remove peer addresses")
    @click.pass_context
    class addr(object):
        """
        Modify peer addresses

        In the future, this might also show addresses, but for now that can be
        done with the "peer show" command.
        """

        def __init__(self, ctx):
            self.organize = (
                ctx.meta.get('Organize', {}).get('magic_instance')
                or organize_dbus_if_active()
            )

        @DualUse.method()
        @click.argument('vk', type=str)
        @click.argument('ip', type=str)
        def add(self, vk, ip):
            """
            Add an address to a peer
            """
            click.echo(Result.from_yaml(self.organize.peer_addr_add(vk, ip)))

        @DualUse.method(name='del')
        @click.argument('vk', type=str)
        @click.argument('ip', type=str)
        def del_(self, vk, ip):
            """
            Delete an address from a peer

            Note: if the peer re-announces the address, it will be re-added. To
            prevent an address from being (re)enabled, you can set it to
            disabled instead of deleting it. This is currently accomplished
            using a command like "peer set $vk IPv4addrs x.x.x.x false"
            """
            click.echo(Result.from_yaml(self.organize.peer_addr_del(vk, ip)))

    @DualUse.method(short_help="Set arbitrary peer properties")
    @click.argument('vk', type=str, **shell_complete_helper(_ac_get_peer_ids))
    @click.argument('path', type=str, nargs=-1)
    @click.argument('value', type=str)
    def set(self, vk, path, value):
        """
        Modify arbitrary peer properties

        This is currently the only way to verify peers, enable/disable them,
        and enable or disable IP addresses.

        In the future, this command should perhaps only be available for
        debugging, and the normal user tasks which it currently performs should
        be handled by other commands.

        This command is *not* able to remove keys from dictionaries; for
        removing IP addresses (instead of disabling them with this command) you
        can use the "vula peer addr del" command. This command is able to add new
        IPs, but that is better done with "vula peer addr add".
        """
        res = Result(json.loads(self.organize.set_peer(vk, path, value)))
        click.echo(res)

    @DualUse.method()
    @click.argument('vk', type=str)
    def remove(self, vk):
        """
        Remove a peer
        """
        res = self.organize.remove_peer(vk)
        res = Result.from_yaml(str(res))
        return res


main = PeerCommands.cli

if __name__ == "__main__":
    import doctest

    print(doctest.testmod())



import gettext

import pkg_resources

from .descriptor_overlay import DescriptorOverlay
from .help_overlay import HelpOverlay
from .peer_details_overlay import PeerDetailsOverlay
from .popupMessage import PopupMessage
from .verification_key_overlay import VerificationKeyOverlay

locale_path = pkg_resources.resource_filename('vula', 'locale')
lang_translations = gettext.translation(
    domain="ui.overlay", localedir=locale_path, fallback=True
)
lang_translations.install()



"""
 *vula-configuration* creates *vula.conf* and enables the *vula*
 WireGuard device with *wg-quick*.
"""

import os
import time
from base64 import b64encode
from logging import Logger, getLogger
from os import geteuid, mkdir, system
from sys import platform

import click
from click.exceptions import Exit
from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.x25519 import X25519PrivateKey

from .csidh import ctidh, ctidh_parameters
from .status import main as StatusCommand

try:
    from dbus import Boolean, Interface, SystemBus
except ImportError:
    pass
from nacl.encoding import Base64Encoder
from nacl.signing import SigningKey

try:
    from systemd import daemon
except ImportError:
    pass

from .common import KeyFile, attrdict
from .constants import _ORGANIZE_KEYS_CONF_FILE, _WG_SERVICES
from .notclick import DualUse


@DualUse.object(
    short_help="Configure system for vula",
    invoke_without_command=True,
)
@click.option(
    "-k",
    "--keys-conf-file",
    default=_ORGANIZE_KEYS_CONF_FILE,
    show_default=True,
    help="YAML configuration file for cryptographic keys",
)
@click.pass_context
class Configure(attrdict):
    def __init__(self, ctx, **kw):
        """
        Configure interface (generating keys if necessary)
        """
        self.update(**kw)
        self.log: Logger = getLogger()
        self.log.debug("Debug level logging enabled")
        self._ctidh = None
        self._ctx = ctx

    def _ensure_root(self):
        if geteuid() != 0:
            self.log.error(
                "unable to configure: needs administrative privileges"
            )
            raise Exit(1)

    def _ensure_linux(self):
        if platform.startswith("linux"):
            if daemon.booted() != 1:
                self.log.error("no systemd: manual configuration required")
                raise Exit(2)

    @DualUse.method()
    def wg_quick_config(self):
        pass

    def _curve25519_keypair_gen(self):
        """
        Generate Curve25519 keypair and return *private* and *public* base64
        encoded values.
        """
        try:
            temp_key = X25519PrivateKey.generate()
        except UnsupportedAlgorithm:
            self.log.debug("Unable to generate Curve25519 keypair")
            raise Exit(3)

        private = b64encode(
            temp_key.private_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PrivateFormat.Raw,
                encryption_algorithm=serialization.NoEncryption(),
            )
        ).decode("utf-8")
        public = b64encode(
            temp_key.public_key().public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw,
            )
        ).decode("utf-8")
        return private, public

    def _ed25519_keypair_gen(self):
        """
        Generate a seed for the Ed25519 verify keypair used by vula.

        """
        self.log.debug("Generating vula Ed25519 keypair...")
        signing_key = SigningKey.generate()
        private = signing_key.encode(encoder=Base64Encoder).decode()
        verify_key = signing_key.verify_key
        public = b64encode(verify_key.encode()).decode()
        return private, public

    def _ctidh_keypair_gen(self):
        if self._ctidh is None:
            self.log.debug("Initializing CTIDH")
            self._ctidh = ctidh(ctidh_parameters)
        self.log.debug("Generating CTIDH keypair")
        sk = self._ctidh.generate_secret_key()
        pk = sk.derive_public_key()
        self.log.debug("CTIDH keypair generated")
        public = b64encode(pk).decode()
        private = b64encode(sk).decode()
        return private, public

    @DualUse.method()
    def generate_or_read_keys(self) -> KeyFile:
        path = self.keys_conf_file
        keys = None
        try:
            keys = KeyFile.from_yaml_file(path)
            os.chmod(path, mode=0o600)
        except FileNotFoundError:
            self.log.info("Keys file not found")
        except Exception as ex:
            backup_path = path + '.' + str(time.time())
            self.log.info(
                "Invalid keys file: (%s); Saving old file as %s",
                ex,
                backup_path,
            )
            os.rename(path, backup_path)
        if keys is None:
            self.log.info("Generating keys...")
            keys = self.genkeys()
            keys.write_yaml_file(path, mode=0o600, autochown=True)
        return keys

    @DualUse.method()
    def genkeys(self):
        keys = KeyFile(
            zip(
                (
                    "pq_ctidhP512_sec_key",
                    "pq_ctidhP512_pub_key",
                    "vk_Ed25519_sec_key",
                    "vk_Ed25519_pub_key",
                    "wg_Curve25519_sec_key",
                    "wg_Curve25519_pub_key",
                ),
                (
                    self._ctidh_keypair_gen()
                    + self._ed25519_keypair_gen()
                    + self._curve25519_keypair_gen()
                ),
            )
        )
        return keys

    def _reconfigure_restart_systemd_services(
        self,
        mode: str = "replace",
        restart: bool = False,
    ):
        """
        Use DBus to check the status of, enable, and restart systemd services.
        """
        system_bus = SystemBus()
        systemd = system_bus.get_object(
            "org.freedesktop.systemd1", "/org/freedesktop/systemd1"
        )
        systemd_manager = Interface(
            systemd, "org.freedesktop.systemd1.Manager"
        )

        if restart:
            for service in _WG_SERVICES:
                systemd_manager.RestartUnit(service, mode)
                self.log.info("Restarted service: %s", service)
        else:
            for service in _WG_SERVICES:
                current_state = systemd_manager.GetUnitFileState(service)
                if current_state == "enabled":
                    self.log.info(
                        "systemd service %s is already enabled", service
                    )
                else:
                    self.log.info(
                        "systemd service %s was %s; enabling it",
                        service,
                        current_state,
                    )

                    systemd_manager.EnableUnitFiles(
                        [service], Boolean(False), Boolean(True)
                    )
                    current_state = systemd_manager.GetUnitFileState(service)
                    self.log.info("%s: %s", service, current_state)
        system_bus.close()

    @DualUse.method()
    def configure_system(self):
        self._ensure_root()
        if platform.startswith("linux"):
            self.log.info("Adding firewall exception with ufw")
            system("ufw allow 5354/udp")
        try:
            mkdir("/etc/wireguard/")
        except FileExistsError:
            pass
        if platform.startswith("linux"):
            system("systemctl daemon-reload")
            system("systemctl restart systemd-sysusers")
            system("systemctl reload dbus")
        if platform.startswith("linux"):
            # XXX: this line is left here to ease upgrades from systems where
            # the wg-quick@vula service is in a degraded state. it should not
            # ever be in a degraded state anymore after this code has run once,
            # now that we've removed the call to "wg-quick up vula". so, after
            # all existing installs have upgraded, we should remove this line:
            system("wg-quick down vula")

    @DualUse.method()
    def nsswitch(self):
        self._ensure_root()
        self.log.info("configuring nsswitch to respect our petname system")
        system(
            "perl -pi -e 'm/vula /s || s/^(hosts:\\s+)/${1}vula /'"
            " /etc/nsswitch.conf"
        )

    @DualUse.method()
    def systemd_restart(self):
        self._ensure_root()
        if platform.startswith("linux"):
            _reconfigure_restart_systemd_services()  # noqa: F821
            _reconfigure_restart_systemd_services(restart=True)  # noqa: F821
            time.sleep(1.5)
        # FIXME: this causes a non-zero exit status sometimes, if the organize
        # service is "activating" instead of "active". the sleep could be
        # increased, or we could perhaps hang around to find out what happened
        # via some dbus event or something?
        self._ctx.invoke(StatusCommand)


main = Configure.cli

if __name__ == "__main__":
    main()



from unittest.mock import mock_open, patch

import pytest

import vula.common


class Testyamlfile:
    def test_from_yaml_file_nonexisting(self):
        with pytest.raises(FileNotFoundError):
            vula.common.yamlfile.from_yaml_file("non-existing-file.yml")

    def test_from_yaml_file(self):
        """
        Note: from_yaml_file uses click to open the file. Click 8.x uses
        builtins.open while click 7.x uses io.open, so we mock both here.
        """
        m = mock_open(read_data="bla:\n  foo: 2")
        with patch("builtins.open", m):
            with patch("io.open", m):
                print(vula.common.yamlfile)
                yaml = vula.common.yamlfile.from_yaml_file(
                    "some_yaml_file.yml"
                )
                assert yaml == {"bla": {"foo": 2}}



"""
Later this file will have some logic to import the right implementation of Sys
for the current operating system. Currently, only pyroute2 on Linux is
supported.
"""

from .sys_pyroute2 import Sys  # noqa: F401



import gettext
import math
from tkinter import Button, Canvas, Frame, Label, PhotoImage
from typing import List

from vula.frontend import DataProvider, PeerType
from vula.frontend.constants import (
    BACKGROUND_COLOR,
    BACKGROUND_COLOR_CARD,
    FONT,
    FONT_SIZE_HEADER_2,
    FONT_SIZE_TEXT_L,
    FONT_SIZE_TEXT_M,
    FONT_SIZE_TEXT_S,
    FONT_SIZE_TEXT_XS,
    IMAGE_BASE_PATH,
    TEXT_COLOR_GREEN,
    TEXT_COLOR_GREY,
    TEXT_COLOR_HEADER_2,
    TEXT_COLOR_PURPLE,
    TEXT_COLOR_WHITE,
    TEXT_COLOR_YELLOW,
)
from vula.frontend.overlay import PeerDetailsOverlay, PopupMessage

_ = gettext.gettext


class Peers(Frame):
    data = DataProvider()

    peer_frames: List[Frame] = []

    # need to save this numbers for updating the GUI,
    # as it takes quite long for dbus to make these changes
    num_peers = 0
    num_peers_after_remove = 0

    # navigation to show more pages with peers if needed
    peer_page = 1
    peers_per_page = 5

    def __init__(self, parent: Frame) -> None:
        Frame.__init__(self, parent)
        self.app = parent

        self.display_header()
        self.display_peers()
        self.display_buttons()
        self.update_loop()

    def display_header(self) -> None:
        self.title_frame = Frame(
            self.app, bg=BACKGROUND_COLOR, width=400, height=40
        )
        title = Canvas(
            self.title_frame,
            bg=BACKGROUND_COLOR,
            height=40,
            width=400,
            bd=0,
            highlightthickness=0,
            relief="ridge",
        )
        title.place(x=0, y=0)
        title.create_text(
            0,
            0,
            anchor="nw",
            text="Peers",
            fill=TEXT_COLOR_HEADER_2,
            font=(FONT, FONT_SIZE_HEADER_2),
        )
        self.title_frame.grid(row=0, column=0, pady=(10, 0), sticky="w")

    def display_buttons(self) -> None:
        self.buttons_frame = Frame(
            self.app, bg=BACKGROUND_COLOR, width=400, height=80
        )
        self.buttons_frame.grid(row=99, column=0, sticky="w", pady=10)

        input_canvas = Canvas(
            self.buttons_frame,
            bg=BACKGROUND_COLOR,
            height=80,
            width=400,
            bd=0,
            highlightthickness=0,
            relief="ridge",
        )
        input_canvas.place(x=0, y=0)

        self.button_image_previous_page = PhotoImage(
            file=IMAGE_BASE_PATH + 'previous.png'
        )
        self.button_previous_page = Button(
            master=self.buttons_frame,
            image=self.button_image_previous_page,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.previous_page(),
            relief="flat",
        )

        self.button_image_next_page = PhotoImage(
            file=IMAGE_BASE_PATH + 'next.png'
        )
        self.button_next_page = Button(
            master=self.buttons_frame,
            image=self.button_image_next_page,
            borderwidth=0,
            highlightthickness=0,
            command=lambda: self.next_page(),
            relief="flat",
        )

        if math.ceil(self.num_peers / self.peers_per_page) > 1:
            self.button_next_page.place(
                x=321.0, y=0.0, width=79.0, height=23.0
            )

    def next_page(self) -> None:
        self.peer_page += 1

        self.button_previous_page.place(
            x=232.0, y=0.0, width=79.0, height=23.0
        )
        if self.peer_page == math.ceil(self.num_peers / self.peers_per_page):
            self.button_next_page.place_forget()

        self.clear_peers()
        self.display_peers()

    def previous_page(self) -> None:
        self.peer_page -= 1

        self.button_next_page.place(x=321.0, y=0.0, width=79.0, height=23.0)
        if self.peer_page == 1:
            self.button_previous_page.place_forget()

        self.clear_peers()
        self.display_peers()

    def display_peers(self) -> None:
        peers = self.data.get_peers()
        self.num_peers = len(peers)
        counter = 1

        if self.num_peers == 0:
            no_peers_label = Label(
                self.app,
                text="No Peers to display",
                bg=BACKGROUND_COLOR,
                fg=TEXT_COLOR_WHITE,
                font=(FONT, FONT_SIZE_TEXT_L),
            )
            no_peers_label.grid(row=counter, column=0, sticky="w", pady=10)

        # Get slice of array for current page
        # [0:5], [5, 10] etc.
        peers_for_page = peers[
            (self.peer_page - 1)
            * self.peers_per_page : (self.peer_page - 1)
            * self.peers_per_page
            + self.peers_per_page
        ]

        for peer in peers_for_page:
            peer_frame = Frame(
                self.app, bg=BACKGROUND_COLOR, width=400, height=70
            )
            peer_frame.grid(row=counter, column=0, sticky="w", pady=10)

            canvas = Canvas(
                peer_frame,
                bg=BACKGROUND_COLOR,
                height=70,
                width=400,
                bd=0,
                highlightthickness=0,
                relief="ridge",
            )

            canvas.place(x=0, y=0)

            self.round_rectangle(
                canvas, 0, 0, 400, 70, r=30, fill=BACKGROUND_COLOR_CARD
            )

            if peer["name"]:
                name = peer["name"]
            else:
                name = peer["other_names"]

            # Peer name
            canvas.create_text(
                20.0,
                10.0,
                anchor="nw",
                text=name,
                fill=TEXT_COLOR_GREEN,
                font=(FONT, FONT_SIZE_TEXT_S),
            )

            # Endpoint IP
            canvas.create_text(
                20.0,
                40.0,
                anchor="nw",
                text=peer["endpoint"],
                fill=TEXT_COLOR_GREY,
                font=(FONT, FONT_SIZE_TEXT_XS),
            )

            # Status labels
            if "enabled" in peer["status"]:
                canvas.create_text(
                    205.0,
                    40.0,
                    anchor="nw",
                    text="enabled",
                    fill=TEXT_COLOR_YELLOW,
                    font=(FONT, FONT_SIZE_TEXT_M),
                )
            if "unpinned" not in peer["status"]:
                canvas.create_text(
                    270.0,
                    40.0,
                    anchor="nw",
                    text="pinned",
                    fill=TEXT_COLOR_PURPLE,
                    font=(FONT, FONT_SIZE_TEXT_M),
                )
            if "unverified" not in peer["status"]:
                canvas.create_text(
                    325.0,
                    40.0,
                    anchor="nw",
                    text="verified",
                    fill=TEXT_COLOR_GREEN,
                    font=(FONT, FONT_SIZE_TEXT_M),
                )

            canvas.bind("<Button-1>", lambda e=_, p=peer: self.open_details(p))

            self.peer_frames.append(peer_frame)
            counter += 1

    def open_details(self, peer: PeerType) -> None:
        popup = PeerDetailsOverlay(self.app, peer)
        result = popup.show()
        if result == "delete":
            self.num_peers_after_remove = self.num_peers - 1
            self.peer_page = 1
            self.clear_peers()
            self.display_peers()
        if (
            result == "pin_and_verify"
            or result == "rename"
            or result == "additional_ip"
        ):
            self.clear_peers()
            self.display_peers()

    def update_loop(self) -> None:
        # function to update the GUI after
        # removing peers (checks number of peers)
        peers = self.data.get_peers()
        if len(peers) > 0:
            if (
                len(peers) == self.num_peers_after_remove
                or len(peers) != self.num_peers
            ):
                try:
                    self.clear_peers()
                    self.display_peers()
                    self.num_peers_after_remove = 0
                except Exception:
                    PopupMessage.showPopupMessage(
                        "Error", "Could not update peers"
                    )

        # check every 5 seconds if number of peers
        # has changed.
        self.after(5000, self.update_loop)

    def clear_peers(self) -> None:
        for frame in self.peer_frames:
            frame.destroy()
        self.app.update()
        self.peer_frames.clear()

    def round_rectangle(
        self,
        canvas: Canvas,
        x: int,
        y: int,
        w: int,
        h: int,
        r=25,
        **kwargs,
    ) -> None:
        xr = x + r
        yr = y + r
        wr = w - r
        hr = h - r
        points = [
            xr,
            y,
            xr,
            y,
            wr,
            y,
            wr,
            y,
            w,
            y,
            w,
            yr,
            w,
            yr,
            w,
            hr,
            w,
            hr,
            w,
            h,
            wr,
            h,
            wr,
            h,
            xr,
            h,
            xr,
            h,
            x,
            h,
            x,
            hr,
            x,
            hr,
            x,
            yr,
            x,
            yr,
            x,
            y,
        ]
        canvas.create_polygon(points, **kwargs, smooth=True)



import gettext
import tkinter as tk
from tkinter import Button, Canvas, Frame, PhotoImage

from vula import common
from vula.frontend import DataProvider
from vula.frontend.constants import (
    BACKGROUND_COLOR,
    FONT,
    FONT_SIZE_HEADER,
    FONT_SIZE_TEXT_L,
    FONT_SIZE_TEXT_XL,
    HEIGHT,
    IMAGE_BASE_PATH,
    TEXT_COLOR_HEADER,
    TEXT_COLOR_WHITE,
    WIDTH,
)
from vula.frontend.overlay import (
    DescriptorOverlay,
    HelpOverlay,
    VerificationKeyOverlay,
)
from vula.frontend.view import Peers, Prefs

_ = gettext.gettext


class App(tk.Tk):
    def __init__(self, *args, **kwargs) -> None:
        data = DataProvider()
        tk.Tk.__init__(self, *args, **kwargs)

        self.geometry("{}x{}".format(WIDTH, HEIGHT))
        self.config(bg=BACKGROUND_COLOR)

        # create all of the main containers
        header_frame = Frame(
            self, bg=BACKGROUND_COLOR, width=1200, height=50, pady=3
        )
        content_frame = Frame(
            self, bg=BACKGROUND_COLOR, width=1200, height=600, padx=3, pady=3
        )
        footer_frame = Frame(
            self, bg=BACKGROUND_COLOR, width=1200, height=150, pady=30, padx=30
        )
        bottom_frame = Frame(
            self, bg=BACKGROUND_COLOR, width=600, height=50, pady=3
        )

        # layout all of the main containers
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)

        header_frame.grid(row=0, sticky="ew")
        content_frame.grid(row=1, sticky="nsew")
        footer_frame.grid(row=2, sticky="e")
        bottom_frame.grid(row=3, sticky="s")

        # create the center widgets
        content_frame.grid_rowconfigure(0, weight=1)
        content_frame.grid_columnconfigure(1, weight=1)

        pref_frame = Frame(
            content_frame,
            bg=BACKGROUND_COLOR,
            width=600,
            height=600,
            padx=3,
            pady=3,
        )
        pref_frame.grid(row=0, column=0, sticky="ns")
        pref_frame.grid_propagate(0)
        self.prefs = Prefs(pref_frame)

        peers_frame = Frame(
            content_frame,
            bg=BACKGROUND_COLOR,
            width=600,
            height=600,
            padx=3,
            pady=3,
        )
        peers_frame.grid(row=0, column=1, sticky="nsew")
        peers_frame.grid_propagate(0)
        self.peers_new = Peers(peers_frame)

        peers_frame.grid_columnconfigure(0, weight=1)

        header = Canvas(
            header_frame,
            bg=BACKGROUND_COLOR,
            height=50,
            width=1200,
            bd=0,
            highlightthickness=0,
            relief="ridge",
        )

        header.place(x=0, y=0)
        header.create_text(
            30.0,
            10.0,
            anchor="nw",
            text="Dashboard",
            fill=TEXT_COLOR_HEADER,
            font=(FONT, FONT_SIZE_HEADER),
        )

        footer_frame.grid_rowconfigure(1, weight=1)
        footer_frame.grid_columnconfigure(1, weight=1)

        vk_label = tk.Label(
            footer_frame,
            text="Verification Key:",
            bg=BACKGROUND_COLOR,
            fg=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_L),
        )
        vk_label.grid(row=0, column=0, sticky="w")

        self.button_image = PhotoImage(file=IMAGE_BASE_PATH + 'show_qr.png')
        vk_button = tk.Button(
            footer_frame,
            text="Show QR",
            image=self.button_image,
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
            command=lambda: self.open_vk_qr_code(),
        )
        vk_button.grid(row=0, column=1)

        desc_label = tk.Label(
            footer_frame,
            text="Descriptor:",
            bg=BACKGROUND_COLOR,
            fg=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_L),
        )
        desc_label.grid(row=1, column=0, sticky="w")

        desc = DescriptorOverlay(self)

        desc_button = tk.Button(
            footer_frame,
            text="Show QR",
            image=self.button_image,
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
            command=lambda: desc.openNewWindow(),
        )
        desc_button.grid(row=1, column=1)

        # Get the status of the different vula processes
        state = data.get_status()

        # Display the status at the bottom
        status_label = tk.Label(
            bottom_frame,
            text=f'Publish: {_(state["publish"])} '
            f'\t Discover: {_(state["discover"])} '
            f'\t Organize: {_(state["organize"])}',
            bg=BACKGROUND_COLOR,
            fg=TEXT_COLOR_WHITE,
            font=(FONT, FONT_SIZE_TEXT_XL),
        )
        status_label.grid(row=0, column=0)

        # Add different command
        self.button_image_rediscover = PhotoImage(
            file=IMAGE_BASE_PATH + 'rediscover.png'
        )
        btn_Rediscover = Button(
            bottom_frame,
            text="Rediscover",
            image=self.button_image_rediscover,
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
            command=lambda: self.rediscover(),
        )
        self.button_image_repair = PhotoImage(
            file=IMAGE_BASE_PATH + 'repair.png'
        )
        btn_Repair = Button(
            bottom_frame,
            text="Repair",
            image=self.button_image_repair,
            command=lambda: self.repair(),
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        self.button_image_gate = PhotoImage(
            file=IMAGE_BASE_PATH + 'release_gateway.png'
        )
        btn_Release_Gateway = Button(
            bottom_frame,
            text="Release Gateway",
            image=self.button_image_gate,
            command=lambda: self.release_gateway(),
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )

        info_Help = HelpOverlay(self)

        self.button_image_help = PhotoImage(file=IMAGE_BASE_PATH + 'help.png')
        btn_Help = Button(
            bottom_frame,
            text="Help",
            image=self.button_image_help,
            command=lambda: info_Help.openNewWindow(),
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        btn_Rediscover.grid(row=0, column=1, pady=(20, 20), padx=(40, 20))
        btn_Repair.grid(row=0, column=2, pady=(20, 20), padx=(20, 20))
        btn_Release_Gateway.grid(row=0, column=3, pady=(20, 20), padx=(20, 20))
        btn_Help.grid(row=0, column=4, pady=(20, 20), padx=(20, 20))

    def rediscover(self) -> None:
        common.organize_dbus_if_active().rediscover()

    def release_gateway(self) -> None:
        common.organize_dbus_if_active().release_gateway()

    def repair(self) -> None:
        common.organize_dbus_if_active().sync(True)

    def open_vk_qr_code(self) -> None:
        VerificationKeyOverlay(self)


def main() -> None:
    try:
        app = App()
        app.title("Vula")
        app.mainloop()
    except RuntimeError as e:
        print(e)



import gettext
import tkinter as tk
from tkinter import (
    Button,
    Canvas,
    Event,
    Frame,
    Label,
    PhotoImage,
    Scrollbar,
    Text,
    Widget,
)
from tkinter.constants import W

from vula.frontend import DataProvider, PrefsType
from vula.frontend.constants import (
    BACKGROUND_COLOR,
    BACKGROUND_COLOR_CARD,
    FONT,
    FONT_SIZE_HEADER_2,
    FONT_SIZE_TEXT_XL,
    IMAGE_BASE_PATH,
    TEXT_COLOR_BLACK,
    TEXT_COLOR_GREEN,
    TEXT_COLOR_GREY,
    TEXT_COLOR_HEADER_2,
    TEXT_COLOR_RED,
    TEXT_COLOR_WHITE,
)
from vula.frontend.overlay import PopupMessage

_ = gettext.gettext


class Prefs(Frame):
    data = DataProvider()

    def __init__(self, frame: Frame) -> None:
        self.show_editable: bool = False
        self.prefs: PrefsType = []
        self.widgets: Widget = {}
        self.frame: Frame = frame

        self.display_header()
        self.display_frames()

        self.get_prefs()
        self.display_prefs()
        self.hide_save_cancel()

    def display_frames(self) -> None:

        # Frames, Canvas and Scrollbars
        self.top_frame = Frame(
            self.frame,
            bg=BACKGROUND_COLOR,
            width=597,
            padx=30,
            pady=10,
            highlightbackground=BACKGROUND_COLOR,
            highlightcolor=BACKGROUND_COLOR,
            highlightthickness=1,
        )

        self.bottom_frame = Frame(
            self.frame,
            bg=BACKGROUND_COLOR,
            width=597,
            height=97,
            padx=30,
            pady=10,
            highlightbackground=BACKGROUND_COLOR,
            highlightcolor=BACKGROUND_COLOR,
            highlightthickness=1,
        )

        self.pref_canvas = Canvas(
            self.top_frame,
            background=BACKGROUND_COLOR_CARD,
            highlightbackground=BACKGROUND_COLOR_CARD,
            highlightcolor=BACKGROUND_COLOR_CARD,
            highlightthickness=1,
        )

        self.yscrollbar = Scrollbar(
            self.top_frame,
            orient="vertical",
            command=self.pref_canvas.yview,
            relief="flat",
        )
        self.pref_content_frame = Frame(
            self.pref_canvas,
            bg=BACKGROUND_COLOR_CARD,
            highlightbackground=BACKGROUND_COLOR_CARD,
            highlightcolor=BACKGROUND_COLOR_CARD,
            highlightthickness=1,
        )

        # Packing and configuring
        self.pref_canvas.pack(side="left", fill="y", expand="yes")

        self.yscrollbar.pack(side="right", fill="y")

        self.pref_canvas.configure(yscrollcommand=self.yscrollbar.set)
        self.pref_canvas.bind(
            '<Configure>',
            lambda e: self.pref_canvas.configure(
                scrollregion=self.pref_content_frame.bbox('all')
            ),
        )

        self.pref_canvas.create_window(
            (0, 0), window=self.pref_content_frame, anchor="nw"
        )

        self.top_frame.pack(
            fill="both", expand="yes", padx=(0, 50), pady=(50, 0), side="top"
        )
        self.bottom_frame.pack(side="left")

        self.top_frame.columnconfigure(0, weight=1)

        # Buttons
        self.button_image_edit = PhotoImage(file=IMAGE_BASE_PATH + 'edit.png')

        self.button_image_save_blue = PhotoImage(
            file=IMAGE_BASE_PATH + 'save_blue.png'
        )

        self.button_image_cancel = PhotoImage(
            file=IMAGE_BASE_PATH + 'cancel.png'
        )

        self.btn_edit = Button(
            master=self.bottom_frame,
            image=self.button_image_edit,
            command=lambda: self.set_editable(),
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        self.btn_save = Button(
            self.bottom_frame,
            image=self.button_image_save_blue,
            command=lambda: self.save_prefs(),
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        self.btn_cancel = Button(
            self.bottom_frame,
            image=self.button_image_cancel,
            command=lambda: self.cancel(),
            borderwidth=0,
            highlightthickness=0,
            relief="sunken",
            background=BACKGROUND_COLOR,
            activebackground=BACKGROUND_COLOR,
            activeforeground=BACKGROUND_COLOR,
        )
        self.btn_edit.pack(side="left", padx=10, pady=10)

    def display_header(self) -> None:
        self.title_frame = Frame(
            self.frame, bg=BACKGROUND_COLOR, padx=30, width=400, height=40
        )
        title = Canvas(
            self.title_frame,
            bg=BACKGROUND_COLOR,
            height=40,
            width=400,
            bd=0,
            highlightthickness=0,
            relief="ridge",
        )
        title.place(x=0, y=0)
        title.create_text(
            0,
            0,
            anchor="nw",
            text="Settings",
            fill=TEXT_COLOR_HEADER_2,
            font=(FONT, FONT_SIZE_HEADER_2),
        )
        self.title_frame.grid(row=0, column=0, pady=(10, 0), sticky="w")

    def get_prefs(self) -> None:
        self.prefs: PrefsType = self.data.get_prefs()

    def toggle(self, event: Event) -> None:
        """
        Toggle bool button value
        """
        if event.widget["text"] == "True":
            event.widget.config(text="False", bg=TEXT_COLOR_RED)
        elif event.widget["text"] == "False":
            event.widget.config(text="True", bg=TEXT_COLOR_GREEN)

    def bool_on_enter(self, event: Event) -> None:
        if event.widget["text"] == "True":
            event.widget.config(bg=TEXT_COLOR_BLACK)
        else:
            event.widget.config(bg=TEXT_COLOR_BLACK)

    def bool_on_leave(self, event: Event) -> None:
        if event.widget["text"] == "True":
            event.widget.config(bg=TEXT_COLOR_GREEN)
        else:
            event.widget.config(bg=TEXT_COLOR_RED)

    def save_prefs(self) -> None:
        for pref, values in self.prefs.items():
            # list based prefs
            if type(values) == list:
                current_list = self.widgets[pref].get("1.0", "end").split()
                for value in current_list:
                    if value not in self.prefs[pref]:
                        res = self.data.add_pref(pref, value)
                        if self.show_error(res) == 1:
                            return
                for value in values:
                    if value not in current_list:
                        res = self.data.remove_pref(pref, value)
                        if self.show_error(res) == 1:
                            return
            # boolean based prefs
            elif type(values) == bool:
                bool_value = str(self.widgets[pref]["text"])
                res = self.data.set_pref(pref, bool_value)
                if self.show_error(res) == 1:
                    return
            # int based prefs
            elif type(values) == int:
                int_value = str(self.widgets[pref].get("1.0", "end"))
                res = self.data.set_pref(pref, int_value)
                if self.show_error(res) == 1:
                    return

        self.get_prefs()
        self.show_editable = False
        self.hide_save_cancel()
        self.update_all()

    def show_error(self, res: str) -> int:
        """
        Show error message if needed and return a status code 0 or 1

        >>> Prefs.show_error(Prefs, 'Test')
        0
        """
        if res and "error:" in res:
            PopupMessage.showPopupMessage(
                "Error", "Preference could not be saved"
            )
            return 1
        else:
            return 0

    def cancel(self) -> None:
        self.get_prefs()
        self.show_editable = False
        self.update_all()
        self.hide_save_cancel()

    def set_editable(self) -> None:
        """
        Set pref variables as editable and update the prefs view
        """
        self.show_editable = True
        self.update_all()
        self.show_save_cancel()

    def update_all(self) -> None:
        """
        Remove everything from the prefs view and display the new state
        """
        for label in self.pref_content_frame.grid_slaves():
            label.grid_forget()
        self.pref_canvas.destroy()
        self.yscrollbar.destroy()
        self.top_frame.destroy()
        self.bottom_frame.destroy()
        self.display_frames()
        self.display_prefs()

    def show_save_cancel(self) -> None:
        self.btn_edit.pack_forget()
        self.btn_cancel.pack(side="left", padx=10, pady=10)
        self.btn_save.pack(side="left", padx=10, pady=10)

    def hide_save_cancel(self) -> None:
        self.btn_edit.pack(side="left", padx=10, pady=10)
        self.btn_save.pack_forget()
        self.btn_cancel.pack_forget()

    def display_prefs(self) -> None:
        # row counter
        counter: int = 1

        # Loop over all preferences and display them
        for pref, value in self.prefs.items():
            # show preference descriptions on the left
            pref_label = Label(
                self.pref_content_frame,
                text=_(str(pref)) + ":",
                font=(FONT, FONT_SIZE_TEXT_XL),
                fg=TEXT_COLOR_WHITE,
                height=1,
                anchor="nw",
                bg=BACKGROUND_COLOR_CARD,
            )
            pref_label.grid(row=counter, column=0, padx=2, pady=2, sticky="nw")

            # list based preferences
            if type(value) == list:

                if self.show_editable:
                    value_text = Text(
                        self.pref_content_frame,
                        height=len(value),
                        width=20,
                        bg=BACKGROUND_COLOR_CARD,
                        fg=TEXT_COLOR_GREY,
                        highlightbackground=BACKGROUND_COLOR_CARD,
                        insertbackground=TEXT_COLOR_WHITE,
                    )
                    self.widgets[pref] = value_text
                    for i in range(len(value)):
                        value_text.insert(tk.END, value[i] + "\n")

                    value_text.grid(
                        row=counter, column=1, padx=1, pady=1, sticky=W
                    )
                    counter += 1

                else:
                    if len(value) == 0:
                        value_label = Label(
                            self.pref_content_frame,
                            text="None",
                            font=(FONT, FONT_SIZE_TEXT_XL),
                            bg=BACKGROUND_COLOR_CARD,
                            fg=TEXT_COLOR_WHITE,
                        )
                        value_label.grid(
                            row=counter, column=1, padx=1, pady=1, sticky=W
                        )
                        counter += 1
                    for i in range(len(value)):
                        value_label = Label(
                            self.pref_content_frame,
                            text=str(value[i]),
                            font=(FONT, FONT_SIZE_TEXT_XL),
                            bg=BACKGROUND_COLOR_CARD,
                            fg=TEXT_COLOR_WHITE,
                        )
                        value_label.grid(
                            row=counter, column=1, padx=1, pady=1, sticky=W
                        )
                        counter += 1

            # bool based preferences (as string)
            elif type(value) == bool:
                if str(value) == "True":
                    color = TEXT_COLOR_GREEN
                    font_color = TEXT_COLOR_GREEN
                    font_color_button = TEXT_COLOR_WHITE

                if str(value) == "False":
                    color = TEXT_COLOR_RED
                    font_color = TEXT_COLOR_RED
                    font_color_button = TEXT_COLOR_WHITE

                if self.show_editable:
                    btn_bool = Button(
                        self.pref_content_frame,
                        text=(str(value)),
                        width=5,
                        height=1,
                        bg=color,
                        fg=font_color_button,
                        highlightbackground=BACKGROUND_COLOR_CARD,
                        highlightcolor=BACKGROUND_COLOR_CARD,
                        borderwidth=0,
                        highlightthickness=0,
                        relief="flat",
                    )
                    btn_bool.widgetName = [pref]
                    self.widgets[pref] = btn_bool
                    btn_bool.grid(
                        row=counter, column=1, padx=1, pady=1, sticky=W
                    )

                    btn_bool.bind("<Leave>", self.bool_on_leave)
                    btn_bool.bind("<Enter>", self.bool_on_enter)
                    btn_bool.bind("<Button-1>", self.toggle)
                else:
                    label = Label(
                        self.pref_content_frame,
                        text=str(value),
                        font=(FONT, FONT_SIZE_TEXT_XL),
                        bg=BACKGROUND_COLOR_CARD,
                        fg=font_color,
                    )
                    label.grid(row=counter, column=1, padx=1, pady=1, sticky=W)

                counter += 1

            # int based preference
            elif type(value) == int:

                if self.show_editable:
                    value_text = Text(
                        self.pref_content_frame,
                        height=1,
                        width=20,
                        bg=BACKGROUND_COLOR_CARD,
                        fg=TEXT_COLOR_GREY,
                        highlightbackground=BACKGROUND_COLOR_CARD,
                        highlightcolor=BACKGROUND_COLOR_CARD,
                        insertbackground=TEXT_COLOR_WHITE,
                    )
                    self.widgets[pref] = value_text
                    value_text.grid(
                        row=counter, column=1, padx=1, pady=1, sticky=W
                    )
                    value_text.insert(tk.END, value)
                else:
                    label = Label(
                        self.pref_content_frame,
                        text=str(value),
                        font=(FONT, FONT_SIZE_TEXT_XL),
                        bg=BACKGROUND_COLOR_CARD,
                        fg=TEXT_COLOR_WHITE,
                    )
                    label.grid(row=counter, column=1, padx=1, pady=1, sticky=W)
                counter += 1


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import gettext

import pkg_resources

from .constants import (
    BACKGROUND_COLOR,
    BACKGROUND_COLOR_CARD,
    BACKGROUND_COLOR_ENTRY,
    BACKGROUND_COLOR_ERROR,
    FONT,
    FONT_SIZE_HEADER,
    FONT_SIZE_HEADER_2,
    FONT_SIZE_TEXT_L,
    FONT_SIZE_TEXT_M,
    FONT_SIZE_TEXT_S,
    FONT_SIZE_TEXT_XL,
    FONT_SIZE_TEXT_XS,
    FONT_SIZE_TEXT_XXL,
    HEIGHT,
    IMAGE_BASE_PATH,
    TEXT_COLOR_BLACK,
    TEXT_COLOR_GREEN,
    TEXT_COLOR_GREY,
    TEXT_COLOR_HEADER,
    TEXT_COLOR_HEADER_2,
    TEXT_COLOR_ORANGE,
    TEXT_COLOR_PURPLE,
    TEXT_COLOR_RED,
    TEXT_COLOR_WHITE,
    TEXT_COLOR_YELLOW,
    WIDTH,
)
from .dataprovider import DataProvider, PeerType, PrefsType, StatusType

locale_path = pkg_resources.resource_filename('vula', 'locale')
lang_translations = gettext.translation(
    domain="ui", localedir=locale_path, fallback=True
)
lang_translations.install()



"""
 vula-discover is a stateless program that prints each WireGuard mDNS
 service and formats the service parameters into a single easy to parse line.
 This program's output is intended to be fed into a daemon that configures
 wireguard peers discovered by vula to configure the local vula
 interface.

 The output of this program may be written to a pipe, a log file, a unix
 socket, the vula-organize dbus Interface for processing descriptors, or any
 other place. It should run with the lowest possible privileges possible. The
 output is not filtered and so adversaries may attempt to inject unreasonable
 hosts such as `127.0.0.1` or other addresses. Care should be taken that only
 addresses for the local network segment are used as WireGuard peers.
"""

from ipaddress import ip_address as ip_addr_parser
from logging import Logger, getLogger
from typing import Optional

import click
import pydbus
from click.exceptions import Exit
from gi.repository import GLib
from pyroute2 import IPRoute
from zeroconf import ServiceBrowser, ServiceInfo, ServiceListener, Zeroconf

from .constants import (
    _DISCOVER_DBUS_NAME,
    _LABEL,
    _ORGANIZE_DBUS_NAME,
    _ORGANIZE_DBUS_PATH,
)
from .peer import Descriptor


class WireGuardServiceListener(ServiceListener):
    """
    *WireGuardServiceListener* is for use with *zeroconf's* *ServiceBrowser*.
    The key=value pairs conform to
    https://tools.ietf.org/html/rfc6763#section-6.4
    """

    def __init__(self, callback):
        super(WireGuardServiceListener, self).__init__()
        self.log: Logger = getLogger()
        self.callback = callback

    def remove_service(
        self, zeroconf: Zeroconf, s_type: str, name: str
    ) -> None:
        """
        *remove_service* does nothing.
        """
        self.log.debug(
            "remove_service called: %s, %s, %s", zeroconf, s_type, name
        )

    def add_service(self, zeroconf: Zeroconf, s_type: str, name: str) -> None:
        """
        When zeroconf discovers a new WireGuard service it calls *add_service*
        which produces a peer descriptor on *stdout*.
        """
        # Typing note:
        # 'Any' works here and while 'Optional[ServiceInfo]' should, it does
        # not unless mypy is called with --no-strict-optional like so:
        #
        #   mypy --ignore-missing-imports  --no-strict-optional discover.py
        info: Optional[ServiceInfo] = zeroconf.get_service_info(s_type, name)
        if info is None:
            return
        data = {k.decode(): v.decode() for k, v in info.properties.items()}

        try:
            desc = Descriptor(data)
        except Exception as ex:
            self.log.debug(
                "discover dropped invalid descriptor: %r (%r)" % (data, ex)
            )
            return

        self.callback(desc)

    def update_service(self, *a, **kw):
        return self.add_service(*a, **kw)


class Discover(object):

    dbus = '''
    <node>
      <interface name='local.vula.discover1.Listen'>
        <method name='listen'>
          <arg type='as' name='ip_addrs' direction='in'/>
        </method>
      </interface>
    </node>
    '''

    def __init__(self):

        self.callbacks = []
        self.browsers = {}
        self.log: Logger = getLogger()

    def callback(self, value):
        for callback in self.callbacks:
            callback(value)

    def listen_on_ip_or_if(self, ip_address, interface):

        """
        Deprecated.

        This is for the cli to accept interface names.
        """

        if interface and ip_address:
            self.log.info("Must pick interface or IP address")
            raise Exit(1)

        ip_addr = None

        if ip_address:
            try:
                ip_addr = str(ip_addr_parser(ip_address))
            except:  # noqa: E722
                self.log.info("Invalid IP address argument")
                raise Exit(3)
            ip_addr: str = ip_address
        elif interface:
            with IPRoute() as ipr:
                index = ipr.link_lookup(ifname=interface)[0]
                a = ipr.get_addr(match=lambda x: x['index'] == index)
                ip_addr: str = dict(a[0]['attrs'])['IFA_ADDRESS']

        if ip_addr:
            self.listen([ip_addr])

    def listen(self, ip_addrs):
        for ip_addr in ip_addrs:
            if ip_addr in self.browsers:
                self.log.info("Not launching a second browser for %r", ip_addr)
                continue
            zeroconf: Zeroconf = Zeroconf(interfaces=[ip_addr])
            self.log.debug("Starting ServiceBrowser for %r", ip_addr)
            browser: ServiceBrowser = ServiceBrowser(
                zeroconf, _LABEL, WireGuardServiceListener(self.callback)
            )
            self.browsers[ip_addr] = (zeroconf, browser)
        for old_ip in list(self.browsers):
            if old_ip not in ip_addrs:
                self.log.info(
                    "Removing old service browser for %r (new ip_addrs=%r)",
                    old_ip,
                    ip_addrs,
                )
                self.browsers[old_ip][1].cancel()
                self.browsers[old_ip][0].close()
                del self.browsers[old_ip]

    def shutdown(self):
        for ip, browser in list(self.browsers.items()):
            del self.browsers[ip]
            browser.cancel()

    def is_alive(self):
        return any(browser.is_alive() for browser in self.browsers.values())

    @classmethod
    def daemon(cls, use_dbus, ip_address, interface):

        """
        This method implements the non-monolithic daemon mode where we run
        Discover in its own process (as we deploy on GNU/systemd).
        """

        loop = GLib.MainLoop()

        discover = cls()

        discover.callbacks.append(lambda d: discover.log.info("%s", d))

        if use_dbus:
            discover.log.debug("dbus enabled")
            system_bus = pydbus.SystemBus()
            process = system_bus.get(
                _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
            ).process_descriptor_string
            discover.callbacks.append(lambda d: process(str(d)))
            system_bus.publish(_DISCOVER_DBUS_NAME, discover)

        discover.listen_on_ip_or_if(ip_address, interface)

        loop.run()


# FIXME: should we shutdown zeroconf objects upon glib shutdown? probably.
#        try:
#            while True:
#                sleep(1)
#        except KeyboardInterrupt:
#            pass
#        finally:
#            discover.shutdown()
#        return 0


@click.command(short_help="Layer 3 mDNS discovery daemon")
@click.option(
    "-d",
    "--dbus/--no-dbus",
    'use_dbus',
    default=True,
    is_flag=True,
    help="use dbus for IPC",
)
@click.option(
    "-I",
    "--ip-address",
    type=str,
    help="bind this IP address instead of automatically choosing which IP "
    "to bind",
)
@click.option(
    #    "-i",
    "--interface",
    type=str,
    help="bind to the primary IP address for the given interface, "
    "automatically choosing which IP to announce",
)
def main(**kwargs):
    Discover.daemon(**kwargs)


if __name__ == "__main__":
    main()



"""
 vula-publish is a program that announces a WireGuard mDNS service as
 informed by Organize over dbus or as controlled by organize in monolith mode.

>>> p = Publish()
>>> type(p.zeroconfs)
<class 'dict'>
>>> type(p.log)
<class 'logging.RootLogger'>

"""

from ipaddress import ip_address
from logging import Logger, getLogger
from platform import node

import click
import pydbus
from gi.repository import GLib
from zeroconf import NonUniqueNameException, ServiceInfo, Zeroconf

from .common import comma_separated_IPs
from .constants import _LABEL, _PUBLISH_DBUS_NAME


class Publish(object):

    dbus = '''
    <node>
      <interface name='local.vula.publish1.Listen'>
        <method name='listen'>
          <arg type='a{sa{ss}}' name='new_announcements' direction='in'/>
        </method>
      </interface>
    </node>
    '''

    def __init__(self):

        self.log: Logger = getLogger()
        self.zeroconfs = {}

    def listen(self, new_announcements):
        # First we remove all old zeroconf listeners that are not in our new
        # instructions
        for ip, zc in list(self.zeroconfs.items()):
            if ip not in new_announcements:
                self.log.info("Removing old service announcement for %r", ip)
                zc.close()
                del self.zeroconfs[ip]
        # Now we add a zeroconf listener for each new IP and ServiceInfo if it
        # is not already existing, else we update the old zc object with the
        # new desc
        for ip_addr, desc in new_announcements.items():
            self.log.debug(
                "Starting mDNS service announcement for %r", ip_addr
            )
            name: str = node() + "." + _LABEL
            service_info: ServiceInfo = ServiceInfo(
                _LABEL,
                name=name,
                addresses=[
                    ip_address(ip).packed for ip in desc['addrs'].split(',')
                ],
                port=int(desc['port']),
                properties=desc,
                server=desc['hostname'],
            )
            zeroconf = self.zeroconfs.get(ip_addr)
            if zeroconf:
                # Do update dance
                self.log.debug("Updating vula service: %s", service_info)
                zeroconf.update_service(service_info)
                self.log.debug("Updating vula service.")

            else:
                zeroconf = self.zeroconfs[ip_addr] = Zeroconf(
                    interfaces=list(map(str, comma_separated_IPs(ip_addr)))
                )
                self.log.debug("Registering vula service: %s", service_info)
                try:
                    zeroconf.register_service(service_info)
                    self.log.debug("Registered vula mDNS publishing service.")
                except NonUniqueNameException:
                    self.log.debug(
                        "Unable to register vula mDNS publishing service."
                    )

    @classmethod
    def daemon(cls):
        """
        This method implements the non-monolithic daemon mode where we run
        Publish in its own process (as we deploy on GNU/systemd).
        """
        loop = GLib.MainLoop()
        publish = cls()
        publish.log.debug("Debug level logging enabled")
        pydbus.SystemBus().publish(_PUBLISH_DBUS_NAME, publish)
        publish.log.debug("dbus enabled")
        loop.run()


@click.command(short_help="Layer 3 mDNS publish daemon")
def main(**kwargs):
    Publish.daemon(**kwargs)


if __name__ == "__main__":
    main()



from typing import List, TypedDict

import pydbus
import yaml

from vula.common import escape_ansi
from vula.constants import _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH


class PeerType(TypedDict):
    name: str
    id: str
    other_names: str
    status: str
    endpoint: str
    allowed_ip: str
    latest_signature: str
    latest_handshake: str
    wg_pubkey: str


class StatusType(TypedDict):
    publish: str
    discover: str
    organize: str


class PrefsType(TypedDict):
    pin_new_peers: bool
    accept_nonlocal: bool
    auto_repair: bool
    subnets_allowed: list
    subnets_forbidden: list
    iface_prefix_allowed: list
    local_domains: list
    ephemeral_mode: bool
    accept_default_route: bool
    record_events: bool
    expire_time: int
    overwrite_unpinned: bool


class DataProvider:
    def get_peers(self) -> List[PeerType]:
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )

        # Get all peer ids from the dbus
        ids = organize.peer_ids("enabled")

        peers: List[PeerType] = []

        # Loop over all ids in the list
        for id in ids:
            # Create empty dict for peer
            peer_dict: PeerType = {
                "name": None,
                "id": None,
                "other_names": None,
                "status": None,
                "endpoint": None,
                "allowed_ip": None,
                "latest_signature": None,
                "latest_handshake": None,
                "wg_pubkey": None,
            }

            # Fill in the data
            peer_raw = organize.show_peer(id)
            peer_clear = escape_ansi(peer_raw)
            peer_lines = peer_clear.split("\n")

            peer_dict["name"] = peer_lines[0].lstrip().split(": ")[1]
            peer_dict["id"] = peer_lines[1].lstrip().split(": ")[1]
            if "other names" in peer_lines[2]:
                peer_dict["other_names"] = (
                    peer_lines[2].lstrip().split(": ")[1]
                )
                peer_dict["status"] = peer_lines[3].lstrip().split(": ")[1]
                peer_dict["endpoint"] = peer_lines[4].lstrip().split(": ")[1]
                peer_dict["allowed_ips"] = (
                    peer_lines[5].lstrip().split(": ")[1]
                )
                peer_dict["latest_signature"] = (
                    peer_lines[6].lstrip().split(": ")[1]
                )
                peer_dict["latest_handshake"] = (
                    peer_lines[7].lstrip().split(": ")[1]
                )
                peer_dict["wg_pubkey"] = peer_lines[8].lstrip().split(": ")[1]
            else:
                peer_dict["status"] = peer_lines[2].lstrip().split(": ")[1]
                peer_dict["endpoint"] = peer_lines[3].lstrip().split(": ")[1]
                peer_dict["allowed_ips"] = (
                    peer_lines[4].lstrip().split(": ")[1]
                )
                peer_dict["latest_signature"] = (
                    peer_lines[5].lstrip().split(": ")[1]
                )
                peer_dict["latest_handshake"] = (
                    peer_lines[6].lstrip().split(": ")[1]
                )
                peer_dict["wg_pubkey"] = peer_lines[7].lstrip().split(": ")[1]

            peers.append(peer_dict)

        return peers

    def get_prefs(self) -> PrefsType:
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )

        # Get the data from the organize dbus
        data = organize.show_prefs()
        items: PrefsType = yaml.safe_load(data)

        return items

    def get_status(self) -> StatusType:
        # Fetch the data from the systemd dbus
        systemd = pydbus.SystemBus().get(".systemd1")

        # Create an empty dict for the status
        status: StatusType = {
            "publish": None,
            "discover": None,
            "organize": None,
        }

        # Define names to consider in the result
        names = ["publish", "discover", "organize"]
        for name in names:
            # Template string for service name
            unit_name = "vula-%s.service" % (name,)
            try:
                unit = pydbus.SystemBus().get(
                    ".systemd1", systemd.GetUnit(unit_name)
                )
                status[name] = unit.ActiveState
            except Exception as ex:
                print(ex)
                return None

        return status

    def our_latest_descriptors(self):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        return organize.our_latest_descriptors()

    def delete_peer(self, peer_vk):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        organize.remove_peer(peer_vk)

    def rename_peer(self, peer_vk, name):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        organize.set_peer(peer_vk, ["petname"], name)

    def pin_and_verify(self, peer_vk, peer_name):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        organize.verify_and_pin_peer(peer_vk, peer_name)

    def add_peer(self, peer_vk, ip):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        organize.peer_addr_add(peer_vk, ip)

    def set_pref(self, pref, value):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        return organize.set_pref(pref, value)

    def add_pref(self, pref, value):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        return organize.add_pref(pref, value)

    def remove_pref(self, pref, value):
        organize = pydbus.SystemBus().get(
            _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
        )
        organize.remove_pref(pref, value)



from warnings import filterwarnings

filterwarnings("ignore")

import os
import time
from glob import glob
from shutil import copy2
from subprocess import check_output
from sys import platform

import setuptools
from setuptools.command.build_ext import build_ext as hookBuild_ext

try:
    from babel.messages import frontend as babel
    compile_catalog = babel.compile_catalog
except ImportError:
    compile_catalog=None
try:
    from stdeb.command.bdist_deb import bdist_deb
    from stdeb.command.sdist_dsc import sdist_dsc

    class sdist_dsc_with_postinst(sdist_dsc):
        def run(self):
            res = super(sdist_dsc_with_postinst, self).run()
            print("Installing vula postinst")
            copy2(
                'misc/python3-vula.postinst',
                self.dist_dir + '/vula-{}/debian/'.format(version),
            )
            return res

except ImportError:
    sdist_dsc = None
    sdist_dsc_with_postinst = None
    bdist_deb = None
try:
    from click_man.commands.man_pages import man_pages
except ImportError:
    man_pages = None

try:
    os.environ['SOURCE_DATE_EPOCH'] = (
        check_output("git log -1 --pretty=%ct", shell=True).decode().strip()
    )
except:
    os.environ['SOURCE_DATE_EPOCH'] = str(int(time.time()))

if os.path.exists('vula/__version__.py'):
    with open("vula/__version__.py", "r") as obj:
        version = str(obj.readline().strip())
        version = version.split('"')[1]

if os.path.exists('requirements.txt'):
    with open("requirements.txt", "r") as obj:
        requirements = obj.read().splitlines()
else:
    # this makes stdeb work
    requirements = []

with open("README.md", "r") as obj:
    long_description = obj.read()

linux_data_files = [
    (
        "/etc/systemd/system/",
        [
            "configs/systemd/vula.slice",
            "configs/systemd/vula.target",
            "configs/systemd/vula-discover.target",
            "configs/systemd/vula-discover-alt.target",
            "configs/systemd/vula-publish.target",
            "configs/systemd/vula-publish-alt.target",
            "configs/systemd/vula-organize-monolithic.target",
            "configs/systemd/vula-discover.service",
            "configs/systemd/vula-discover-alt.service",
            "configs/systemd/vula-publish.service",
            "configs/systemd/vula-publish-alt.service",
            "configs/systemd/vula-organize.service",
            "configs/systemd/vula-organize-monolithic.service",
        ],
    ),
    (
        "/etc/dbus-1/system.d/",
        ['configs/dbus/local.vula.services.conf'],
    ),
    (
        "/usr/share/applications",
        [
            "misc/linux-desktop/vula.desktop",
            "misc/linux-desktop/vula-tray.desktop",
        ]
    ),
    (
        "/usr/share/icons",
        [
            "misc/linux-desktop/vula_gui_icon.png"
        ]
    ),
        (
        "/usr/share/icons/vula",
        [
            "misc/images/add_ip_to_peer_entry.png",
            "misc/images/clipboard.png",
            "misc/images/delete.png",
            "misc/images/edit_name_entry.png",
            "misc/images/pin_and_verify.png",
            "misc/images/save.png",
            "misc/images/show_qr.png",
            "misc/images/next.png",
            "misc/images/previous.png",
            "misc/images/edit.png",
            "misc/images/save_blue.png",
            "misc/images/cancel.png",
            "misc/images/help.png",
            "misc/images/repair.png",
            "misc/images/rediscover.png",
            "misc/images/release_gateway.png",
        ]
    ),
    (
        "/usr/share/dbus-1/system-services/",
        [
            'configs/dbus/local.vula.organize.service',
            'configs/dbus/local.vula.publish.service',
            'configs/dbus/local.vula.discover.service',
            'configs/dbus/local.vula.publishalt.service',
            'configs/dbus/local.vula.discoveralt.service',
        ],
    ),
    (
        "/usr/share/polkit-1/actions/",
        ['configs/polkit/local.vula.organize.Debug.policy'],
    ),
    ("/usr/lib/sysusers.d/", ['configs/sysusers.d/vula.conf']),
    (
        "/usr/share/man/man1/",
        glob('man/vula*1'),
    ),
    (
        "",
        ["misc/python3-vula.postinst"],
    ),
]

# The locations for files on macOS. This differs from Linux in
# that the files are stored in usr/local instead of usr/share.
macos_data_files = [
    (
        "/usr/local/share/dbus-1/",
        ['configs/dbus/local.vula.services.conf'],
    ),
    (
        "/usr/local/dbus-1/system-services/",
        [
            'configs/dbus/local.vula.organize.service',
            'configs/dbus/local.vula.publish.service',
            'configs/dbus/local.vula.discover.service',
            'configs/dbus/local.vula.publish-alt.service',
            'configs/dbus/local.vula.discover-alt.service',

        ],
    ),
    ("/usr/local/lib/sysusers.d/", ['configs/sysusers.d/vula.conf']),
    (
        "/usr/local/man/man1/",
        glob('man/vula*1'),
    ),
    (
        "",
        ["misc/python3-vula.postinst"],
    ),
]

# Uses the macOS specific paths for Darwin (macOS) systems
if platform == "darwin":
    our_data_files = macos_data_files
else:
    our_data_files = linux_data_files

if platform.startswith("openbsd"):
    our_data_files = []


class print_version(hookBuild_ext):
    def run(self):
        print(version)


setuptools.setup(
    name="vula",
    version=version,
    author="Vula Authors",
    author_email="git@vula.link",
    description=("Automatic local network encryption"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://codeberg.org/vula/vula",
    packages=setuptools.find_packages(),
    keywords="WireGuard, mDNS, encryption, post-quantum, local-area network, CTIDH, CSIDH, Curve25519",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    entry_points={
        "console_scripts": [
            "vula=vula.__main__:main",
        ]
    },
    install_requires=requirements,
    data_files=our_data_files,
    package_data={'vula': ['vula/locale/*/LC_MESSAGES/*.mo']},
    include_package_data=True,
    zip_safe=False,
    tests_require=["pytest"],
    cmdclass=dict(
        compile_catalog=compile_catalog,
        bdist_deb=bdist_deb,
        sdist_dsc=sdist_dsc_with_postinst,
        man_pages=man_pages,
        version=print_version,
    ),
)



"""
 vula-verify is a stateless program that generates and reads QR codes to
 verify a peer or a set of vula peers.  This program's output is intended to
 be fed into the vula-organize daemon. It should authenticate an already
 known peer or peers and sets a bit to keep state that it has verified them.
 The QR code also includes a PSK and later will use CTIDH to automatically set
 a PSK on a pair-wise basis.

 The output of this program may be written to a pipe, a log file, a unix
 socket, or any other place. It should run with the lowest possible privileges
 possible. The output is not filtered and so adversaries may attempt to inject
 unreasonable data. Care should be taken that the data should only be used
 after it has been verified.
"""

try:
    import cv2
    from pyzbar.pyzbar import decode
except ImportError:
    zbar = None
    cv2 = None
    decode = None

try:
    import qrcode
except ImportError:
    qrcode = None

import hashlib
import json
import sys

import click
import pydbus
import yaml
from click.exceptions import Exit
from gi.repository import GLib

from .common import escape_ansi
from .constants import _ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH
from .engine import Result
from .notclick import DualUse, bold, green, red
from .peer import Descriptor
from .verify_audio import VerifyAudio


@DualUse.object(
    short_help="Verify and share peer verification information",
    invoke_without_command=False,
)
@click.pass_context
class VerifyCommands(object):
    def __init__(self, ctx):
        organize = ctx.meta.get('Organize', {}).get('magic_instance')

        if not organize:
            bus = pydbus.SystemBus()
            organize = bus.get(_ORGANIZE_DBUS_NAME, _ORGANIZE_DBUS_PATH)

        self.organize = organize
        self.my_descriptors = {
            ip: Descriptor(d)
            for ip, d in json.loads(
                self.organize.our_latest_descriptors()
            ).items()
        }
        (self.vk,) = set(d.vk for d in self.my_descriptors.values())

    @DualUse.method()
    def my_vk(self):
        click.echo(green(bold("Your VK is: ")) + str(self.vk))
        qr = qrcode.QRCode()
        qr.add_data(data="local.vula:vk:" + str(self.vk))
        qr.print_ascii()

    @DualUse.method()
    def my_descriptor(self):
        for ip, desc in self.my_descriptors.items():
            click.echo(green(bold("Descriptor for {}: ".format(ip))))
            qr = qrcode.QRCode()
            qr.add_data(data="local.vula:desc:" + str(desc))
            qr.print_ascii()
            click.echo(repr(str(desc)))

    @DualUse.method()
    @click.argument('name', type=str)
    def against(self, name):
        # take name vk and vk (self), hash with sha256
        pass

    @DualUse.method()
    @click.option('-w', '--width', default=640, show_default=True)
    @click.option('-h', '--height', default=480, show_default=True)
    @click.option('-c', '--camera', default=0, show_default=True)
    @click.option(
        '-d', '--debug', default=False, is_flag=True, show_default=True
    )
    @click.argument('hostname', type=str, required=True)
    def scan(self, width, height, camera, hostname, debug):
        """
        We expect a string object that roughly looks like the following three
        things:


            local.vula:desc:<descriptor base64 representation>
            local.vula:vk:<vk base64 representation>
            local.vula:aead:<aead ciphertext base64 representation>

        The first part of the string is conformant to RFC 1738, Section 2.1.
        which describes "The main parts of URLs".
        """
        res = None
        done = False
        data = None
        v = cv2.VideoCapture(camera, cv2.CAP_V4L)
        v.set(cv2.CAP_PROP_FRAME_WIDTH, width)
        v.set(cv2.CAP_PROP_FRAME_HEIGHT, height)
        while v.isOpened():
            ret, img = v.read()
            if not ret:
                break
            cv2.imshow("scan vula qrcode", img)
            k = cv2.waitKey(1)
            if k == ord('q'):
                break
            res = decode(img)
            if res:
                if debug:
                    click.echo(res)
                for result in res:
                    if result.type == 'QRCODE':
                        if result.data.decode('utf-8').startswith(
                            'local.vula:'
                        ):
                            data = result.data.decode('utf-8')
                            done = True
                            break
            if done:
                break
        if not data:
            exit(1)
        data = data.split(':', 1)[1]
        sub_type, data = data.split(':', 1)
        if sub_type == "desc":
            res = self.organize.process_descriptor_string(data)
            res = Result(yaml.safe_load(res))
            click.echo(res)
        elif sub_type == "vk":
            vk = self.organize.get_vk_by_name(hostname)
            if vk == data:
                res = self.organize.verify_and_pin_peer(vk, hostname)
                res = Result(yaml.safe_load(res))
                click.echo(res)
                if res.error is not None:
                    raise Exception(res.error)
            else:
                click.echo("keys are for the wrong DeLorean")
                raise Exit(1)
        else:
            click.echo("unknown qrcode subtype")
            raise Exit(1)
        if res.error is not None:
            click.echo(res)
            raise Exit(1)
        else:
            if debug:
                click.echo(res)
            raise Exit(0)

    @DualUse.method()
    @click.option(
        '-v', '--verbose', default=False, is_flag=True, show_default=True
    )
    def speak(self, verbose):
        vk = str(self.vk)
        if verbose:
            click.echo(f"Sending vk: {vk}")
        VerifyAudio(verbose).send_verification_key(vk)
        if verbose:
            click.echo("Done speaking.")

    @DualUse.method()
    @click.argument('hostname', type=str, required=True)
    @click.option(
        '-v', '--verbose', default=False, is_flag=True, show_default=True
    )
    def listen(self, hostname, verbose):
        try:
            known_vk = self.organize.get_vk_by_name(hostname)
        except GLib.Error:
            click.echo(
                f"hostname {hostname} unknown, "
                f"can only verify previously discovered hosts"
            )
            sys.exit(6)
        click.echo(green(bold('Listening ... ')) + 'Press Ctrl+C to stop')
        received_vk = VerifyAudio(verbose).receive_verification_key()
        if received_vk is not None:
            sanitized_vk: str = escape_ansi(received_vk)
            if sanitized_vk != received_vk:
                click.echo(
                    "The received verification key had to be sanitized, "
                    "this is a possible injection attack!"
                )
                raise Exit(1)
            if verbose:
                click.echo(f"Received VK: {sanitized_vk}")
            if (
                hashlib.sha256(known_vk.encode('utf-8')).hexdigest()
                == received_vk
            ):
                res = self.organize.verify_and_pin_peer(known_vk, hostname)
                res = Result(yaml.safe_load(res))
                click.echo(res)
                if res.error is not None:
                    raise Exception(res.error)
            else:
                click.echo(
                    red(
                        bold(
                            f"The received vk: {sanitized_vk}"
                            f" and known vk: {known_vk} "
                            f"of hostname {hostname} do not match."
                        )
                    )
                )
                if verbose:
                    print("keys are for the wrong DeLorean")
                raise Exit(1)


main = VerifyCommands.cli
if __name__ == "__main__":
    main()



import tempfile
import tkinter as tk
from tkinter import ttk

from vula.frontend.constants import BACKGROUND_COLOR, TEXT_COLOR_WHITE

try:
    import qrcode
except ImportError:
    qrcode = None


class QRCodeLabel(ttk.Label):
    def __init__(self, parent: tk.Frame, qr_data: str, resize: int) -> None:
        with tempfile.NamedTemporaryFile(
            prefix="vula-ui", suffix="qr-code"
        ) as tmp_file:
            tmp_file_name = tmp_file.name
            super().__init__(parent)
            qr = qrcode.QRCode()
            qr.add_data(data=qr_data)
            img = qr.make_image(
                fill_color=TEXT_COLOR_WHITE, back_color=BACKGROUND_COLOR
            )
            img.save(tmp_file_name)
            self.image = tk.PhotoImage(file=tmp_file_name)
            self.image = self.image.subsample(resize, resize)
            self.configure(image=self.image)
